/*
 * Copyright 2008, James William Pye. http://jwp.name
 *//*
 * See the included LICENSE file for conditions of use.
 *
 * BSD Licensed
 */

-- Most of these functions take single identifier parameters to reference
-- tables. However, in cases where the table needs to be referenced in the
-- information_schema, the table_schema and table_name must be distinct
-- parameters.

CREATE OR REPLACE FUNCTION
key_update(
 using_table_ident text,
 target_table_schema text,
 target_table_name text,
 except_columns text[],
 key_columns text[]
)
RETURNS text LANGUAGE SQL STABLE
AS $func$
 SELECT 'UPDATE '
  || qname($2, $3) || ' AS MOLD'
  -- Set all the non-key columns to the "MNEW" values.
  || ' SET '
  || array_to_string(ARRAY(
       SELECT
        quote_ident(column_name) || ' = MNEW.' || quote_ident(column_name)
       FROM information_schema.columns
       WHERE table_schema = $2 AND table_name = $3
        AND (NOT column_name = ANY ($4) OR $4 IS NULL)
        AND (NOT column_name = ANY ($5))
       ORDER BY ordinal_position),
      ', ' -- join by string
     )
  -- From the using_table_ident (MNEW)
  || ' FROM ' || quote_ident($1) || ' AS MNEW WHERE ('
  -- WHERE clause requiring that the keys be consistent.
  || array_to_string(
      ARRAY(
       SELECT 'MOLD.' || quote_ident(column_name) || ' = MNEW.' || quote_ident(column_name)
       FROM (
        -- unroll array of key column names
        SELECT $5[i] AS column_name FROM generate_series(1, array_upper($5, 1)) AS g(i)
       ) AS key_columns (column_name)
       UNION ALL SELECT 'TRUE'::varchar
      ),
      ' AND ' -- join by string
     ) || ')'
$func$;

COMMENT ON FUNCTION key_update(text,text,text,text[],text[]) IS
'Given a source and a destination, produce an update statement that '
'copies the non-key columns to the destinaton when the keys matchs, excepting '
'the specified list of columns to not copy over'
'(using, target_table_schema, target_table_name, exceptions[], key_columns). '
'"MOLD" refers to the table being updated; the "old" row, and "MNEW" refers '
'to the new row.';

--

CREATE OR REPLACE FUNCTION key_update(
 using_table_ident text, target_table_schema text, target_table_name text, except_columns text[]
) RETURNS text LANGUAGE SQL STABLE AS
'SELECT key_update($1, $2, $3, $4, ARRAY(SELECT table_key_columns($2, $3)))';

COMMENT ON FUNCTION key_update(text,text,text,text[]) IS
'key_update with a default of the destination table''s primary key';

--

CREATE OR REPLACE FUNCTION
delete_using(target_table_ident text, using_table_ident text, using_columns text[])
RETURNS text LANGUAGE SQL STABLE AS
$func$
 SELECT
  'DELETE FROM ' || $1 || ' AS target__ '
  || ' USING ' || $2 || ' AS using__ WHERE '
  || array_to_string(ARRAY(
	SELECT
	 'target__.' || quote_ident(COALESCE($3[i][1],$3[i])) || ' = ' ||
	 'using__.' || quote_ident(COALESCE($3[i][2],$3[i]))
	 FROM generate_series(1, array_upper($3, 1)) g(i)),
	' AND ' -- join string
  )
$func$;

COMMENT ON FUNCTION delete_using(text,text,text[]) IS
'Based on the given columns, yield a delete statement '
'that removes entries in the source table that already exist in the '
'destination table.';

--

CREATE OR REPLACE FUNCTION
delete_exact(
 target_ident text,
 using_table_schema text, using_table_name text,
 except_columns text[]
)
RETURNS text LANGUAGE SQL STABLE AS
$func$
 SELECT
    -- Using DELETE USING
    'DELETE FROM ' || $1 || ' AS target__ USING '
    || qname($2, $3) || ' AS using__ WHERE ' || array_to_string(ARRAY(
		  SELECT 
		   '(' || CASE WHEN is_nullable = 'YES' THEN
			  'COALESCE(' || qname('using__', column_name) || ', '
			  || qname('target__', column_name) || ') IS NULL OR '
			  ELSE '' END
		   || '(' || qname('using__', column_name)
		   || ' = ' || qname('target__', column_name) || '))' AS col_eq
		  FROM
			information_schema.columns
		  WHERE table_schema = $2 AND table_name = $3
		   AND (NOT column_name = ANY ($4) OR $4 IS NULL)
		  ORDER BY ordinal_position),
		 ' AND ' -- array_to_string join string.
		)
$func$;

COMMENT ON FUNCTION delete_exact(text,text,text,text[]) IS
'Based on the columns in the destination table with exceptions for the '
'specified exception columns, yield a delete statement '
'that removes entries in the source table that already exist in the '
'destination table.';

--

CREATE OR REPLACE FUNCTION
delete_exact(
 target_table_ident text,
 using_table_schema text,
 using_table_name text
)
RETURNS text LANGUAGE SQL STABLE AS
'SELECT delete_exact($1, $2, $3, NULL::text[])';

COMMENT ON FUNCTION delete_exact(text,text,text) IS
'Based on the columns in the destination table, yield a delete statement '
'that removes entries in the source table that already exist in the '
'destination table.';

--

CREATE OR REPLACE FUNCTION
delete_exact(target_table_ident text, using_table_name text)
RETURNS text LANGUAGE SQL STABLE AS
'SELECT delete_exact($1, table_schema($2), $2, NULL::text[])';

COMMENT ON FUNCTION delete_exact(text,text) IS
'Wraps delete_exact(text,text,text), but resolves the using_table_schema from '
'the search_path using ``table_schema(table_name)``.';
