/*
 * Copyright 2008, James William Pye. http://jwp.name
 *//*
 * See the included LICENSE file for conditions of use.
 *
 * BSD Licensed
 */
CREATE OR REPLACE FUNCTION
merge_by_key(
 IN dest_schema_name text,
 IN dest_table_name text,
 IN except_columns text[],
 IN merge_where_clause text,
 IN source_query text,

 OUT start_time timestamp,
 OUT end_time timestamp,

 OUT create_count bigint,
 OUT create_time interval,

 OUT update_using_count bigint,
 OUT update_using_time interval,

 OUT delete_using_count bigint,
 OUT delete_using_time interval,

 OUT insert_count bigint,
 OUT insert_time interval
)
LANGUAGE plpgsql AS
$func$
DECLARE
 tmp_name text;
 cols text;

 op_start_time timestamp;

 dst_schema_name ALIAS FOR $1;
 dst_table_name ALIAS FOR $2;
 in_except_columns ALIAS FOR $3;
 update_where_clause ALIAS FOR $4;
 select_query ALIAS FOR $5;

 dml_query text;
 create_cmp_count bigint;

 except_columns text[];
BEGIN
 -- Check to make sure the table exists.
 -- It will need read and write on the target.
 EXECUTE 'SELECT * FROM '
  || (CASE
   WHEN dst_schema_name IS NULL
   THEN quote_ident(dst_table_name)
   ELSE qname(dst_schema_name, dst_table_name)
   END) || ' LIMIT 0';

 -- Check for the primary key.
 IF (table_key_ordinals(dst_schema_name, dst_table_name))[1] IS NULL
 THEN
  RAISE EXCEPTION 'merge_by_key target table "%.%" does not have a primary key', dst_schema_name, dst_table_name;
 END IF;

 start_time := timeofday()::timestamp;

 -- Materialize the result into the temporary table.
 -- This is done explicitly rather than taking a table as the `delete exact`
 -- step of the process will remove rows the source.
 --
 -- Also, the given query is expected to produce distinct results,
 -- so if the query were to be flawed this gives merge_by_key a way
 -- to identify the offending rows and provide more useful information
 -- in the exception.
 op_start_time := timeofday()::timestamp;
 tmp_name := dst_table_name || '_merging__tmp_';

 -- Create the temporary source table
 EXECUTE 'CREATE TEMP TABLE ' || tmp_name || ' AS ' || select_query
  || (CASE
   WHEN version() ~* 'greenplum' THEN 
   ' DISTRIBUTED BY ' || array_to_string(ARRAY(
    SELECT quote_ident(
     table_distribution_columns(dst_schema_name, dst_table_name)
    )
	), ',')
	ELSE '' END);

 -- Get the number of rows in the source
 create_time := timeofday()::timestamp - op_start_time;
 FOR create_count IN EXECUTE 'SELECT count(*) AS create_count FROM ' || tmp_name
 LOOP
  
 END LOOP;

 -- Get the columns in the destination that do not exist in the source
 except_columns := ARRAY(
  SELECT column_name::text FROM information_schema.columns
   WHERE table_schema = dst_schema_name AND table_name = dst_table_name
  EXCEPT SELECT column_name::text FROM information_schema.columns
   WHERE table_schema = table_schema(tmp_name) AND table_name = tmp_name
 );
 IF except_columns IS NOT NULL AND array_upper(except_columns, 1) > 0
 THEN
 	RAISE NOTICE 'ignoring absent columns for merge, %', except_columns;
 	except_columns := except_columns || COALESCE(in_except_columns, '{}'::text[]);
 ELSE
 	except_columns := in_except_columns;
 END IF;

 -- Update those PK matches in the target relation that
 -- the materialized set contains. Subsequently, delete them from the tmp.
 dml_query := key_update(tmp_name,
 	dst_schema_name, dst_table_name, except_columns)
	|| COALESCE(' AND ' || update_where_clause, '');
 op_start_time := timeofday()::timestamp;
 EXECUTE dml_query;
 GET DIAGNOSTICS update_using_count = ROW_COUNT;
 update_using_time := timeofday()::timestamp - op_start_time;

 -- The matching keys having been updated, now delete them for the insert.
 dml_query := delete_using(
  quote_ident(tmp_name), qname(dst_schema_name, dst_table_name),
  ARRAY(SELECT table_key_columns(dst_schema_name, dst_table_name))
 );
 op_start_time := timeofday()::timestamp;
 EXECUTE dml_query;
 GET DIAGNOSTICS delete_using_count = ROW_COUNT;
 delete_using_time := timeofday()::timestamp - op_start_time;

 IF delete_using_count < update_using_count
 THEN
  RAISE WARNING 'fewer rows deleted(%) than updated(%) by the key',
   delete_using_count, update_using_count;
 END IF;

 cols := array_to_string(
  ARRAY(
   SELECT quote_ident(column_name) FROM information_schema.columns
   WHERE table_schema = dst_schema_name AND table_name = dst_table_name
	 AND (except_columns IS NULL OR NOT column_name = ANY (except_columns))
   ORDER BY ordinal_position
  ),
  ','
 );

 -- Inserting rows that did not exist in target
 dml_query := 'INSERT INTO '
   || qname(dst_schema_name, dst_table_name)
   || '(' || cols || ')'
   || ' SELECT ' || cols || ' FROM ' || tmp_name;
 op_start_time := timeofday()::timestamp;
 EXECUTE dml_query;
 GET DIAGNOSTICS insert_count = ROW_COUNT;
 insert_time := timeofday()::timestamp - op_start_time;
 create_cmp_count := insert_count + delete_using_count;
 IF create_count != create_cmp_count
 THEN
  RAISE WARNING 'processing inconsistency: % rows created, but % processed',
   create_count, create_cmp_count;
 END IF;

 EXECUTE 'DROP TABLE ' || tmp_name;
 end_time := timeofday()::timestamp;
END;
$func$;

COMMENT ON FUNCTION merge_by_key(text,text,text[],text,text) IS
'Merge the result set of the given query into the given destination table '
'using the primary key on the destination table. '
'When removing exact duplicates and inserting, ignore the columns named in '
'the array. '
''
'Yield information about the timing and counts of the operation.';

CREATE OR REPLACE FUNCTION merge_by_key(
 text, text, text[], text,

 OUT start_time timestamp,
 OUT end_time timestamp,

 OUT create_count bigint,
 OUT create_time interval,

 OUT update_using_count bigint,
 OUT update_using_time interval,

 OUT delete_using_count bigint,
 OUT delete_using_time interval,

 OUT insert_count bigint,
 OUT insert_time interval
)
LANGUAGE SQL AS
'SELECT * FROM merge_by_key($1, $2, $3, NULL::text, $4)';

COMMENT ON FUNCTION merge_by_key(text,text,text[],text) IS
'merge_by_key without specifying the update clause';

CREATE OR REPLACE FUNCTION merge_by_key(
 text, text, text, text,

 OUT start_time timestamp,
 OUT end_time timestamp,

 OUT create_count bigint,
 OUT create_time interval,

 OUT update_using_count bigint,
 OUT update_using_time interval,

 OUT delete_using_count bigint,
 OUT delete_using_time interval,

 OUT insert_count bigint,
 OUT insert_time interval
)
LANGUAGE SQL AS
'SELECT * FROM merge_by_key($1, $2, NULL::text[], $3, $4)';

COMMENT ON FUNCTION merge_by_key(text,text,text,text) IS
'merge_by_key without specifying the exceptions';

CREATE OR REPLACE FUNCTION merge_by_key(
 text, text,

 OUT start_time timestamp,
 OUT end_time timestamp,

 OUT create_count bigint,
 OUT create_time interval,

 OUT update_using_count bigint,
 OUT update_using_time interval,

 OUT delete_using_count bigint,
 OUT delete_using_time interval,

 OUT insert_count bigint,
 OUT insert_time interval
)
LANGUAGE SQL AS
'SELECT * FROM merge_by_key(table_schema($1), $1, NULL::text[], $2)';

CREATE OR REPLACE FUNCTION merge_by_key(text, text[], text,
 OUT start_time timestamp,
 OUT end_time timestamp,

 OUT create_count bigint,
 OUT create_time interval,

 OUT update_using_count bigint,
 OUT update_using_time interval,

 OUT delete_using_count bigint,
 OUT delete_using_time interval,

 OUT insert_count bigint,
 OUT insert_time interval
)
LANGUAGE SQL AS
'SELECT * FROM merge_by_key(table_schema($1), $1, $2, $3)';

CREATE OR REPLACE FUNCTION merge_by_key(text, text, text,
 OUT start_time timestamp,
 OUT end_time timestamp,

 OUT create_count bigint,
 OUT create_time interval,

 OUT update_using_count bigint,
 OUT update_using_time interval,

 OUT delete_using_count bigint,
 OUT delete_using_time interval,

 OUT insert_count bigint,
 OUT insert_time interval
)
LANGUAGE SQL AS
'SELECT * FROM merge_by_key(table_schema($1), $1, NULL::text[], $2, $3)';

COMMENT ON FUNCTION merge_by_key(text,text,text) IS
'merge_by_key without specifying exception columns and when clause';

CREATE OR REPLACE FUNCTION merge_by_key(text, text[], text, text,
 OUT start_time timestamp,
 OUT end_time timestamp,

 OUT create_count bigint,
 OUT create_time interval,

 OUT update_using_count bigint,
 OUT update_using_time interval,

 OUT delete_using_count bigint,
 OUT delete_using_time interval,

 OUT insert_count bigint,
 OUT insert_time interval
)
LANGUAGE SQL AS
'SELECT * FROM merge_by_key(table_schema($1), $1, $2, $3, $4)';

CREATE OR REPLACE FUNCTION merge_by_key_version()
RETURNS text IMMUTABLE LANGUAGE SQL AS 'SELECT ''1.0''::text';
