/*
 * Copyright 2008, James William Pye. http://jwp.name
 *//*
 * See the included LICENSE file for conditions of use.
 *
 * BSD Licensed
 */
CREATE OR REPLACE FUNCTION
table_columns(schema_name text, table_name text)
RETURNS SETOF text LANGUAGE SQL STABLE AS
$func$
 SELECT
  column_name::text
 FROM
  information_schema.columns
 WHERE table_schema = $1 AND table_name = $2
 ORDER BY ordinal_position
$func$;

COMMENT ON FUNCTION table_columns(text,text) IS
'Return the table''s columns in the order of oridinal_position.';

--

CREATE OR REPLACE FUNCTION
table_columns(
 schema_name text, table_name text, column_positions smallint[]
)
RETURNS SETOF text LANGUAGE SQL STABLE STRICT AS
$func$
SELECT
 column_name
FROM (
  SELECT
   i, ($3[i])::int AS ordinal_position
  FROM generate_series(1, array_upper($3, 1)) g(i)
 ) AS dist
 JOIN
  (SELECT
    column_name::text,
    ordinal_position::int
   FROM
    information_schema.columns
   WHERE table_schema = $1 AND table_name = $2
  ) AS info
  USING (ordinal_position)
ORDER BY i
$func$;

COMMENT ON FUNCTION table_columns(text, text, smallint[]) IS
'Return the table''s columns in the order designated by ``column_positions`` '
'array';

--

CREATE OR REPLACE FUNCTION
table_key_ordinals(table_schema text, table_name text)
RETURNS smallint[] LANGUAGE SQL STABLE AS
$func$
SELECT ARRAY(
	SELECT
	 ordinal_position::smallint
	FROM information_schema.key_column_usage
	WHERE table_schema = $1 AND table_name = $2
	ORDER BY position_in_unique_constraint ASC
);
$func$;

COMMENT ON FUNCTION table_key_ordinals(text,text) IS
'return an array of the table''s key column ordinals';

--

CREATE OR REPLACE FUNCTION
table_key_columns(schema_name text, table_name text)
RETURNS SETOF text LANGUAGE SQL STABLE AS
'SELECT table_columns($1, $2, table_key_ordinals($1, $2))';

COMMENT ON FUNCTION table_key_columns(text,text) IS
'return a set of the key column names of the table';

--

CREATE OR REPLACE FUNCTION
table_distribution_ordinals(table_schema text, table_name text)
RETURNS smallint[] STABLE STRICT LANGUAGE plpgsql AS
$func$
DECLARE
 r smallint[];
BEGIN
 r := NULL;

 IF version() ~* 'greenplum'
 THEN
    -- Greenplum 3.0
    SELECT attrnums INTO r FROM
     pg_namespace JOIN pg_class
      ON (relnamespace = pg_namespace.oid AND relname = $2)
     JOIN gp_distribution_policy ON (localoid = pg_class.oid)
    WHERE nspname = $1;
 END IF;

 RETURN r;
END
$func$;

COMMENT ON FUNCTION table_distribution_ordinals(text, text) IS
'';

--

CREATE OR REPLACE FUNCTION
table_distribution_columns(table_schema text, table_name text)
RETURNS SETOF text STABLE STRICT LANGUAGE SQL AS
'SELECT table_columns($1, $2, table_distribution_ordinals($1, $2))';

COMMENT ON FUNCTION table_distribution_columns(text, text) IS
'Columns that are used for partitioning. Used for proper temporary table '
'creation.';

CREATE OR REPLACE FUNCTION
table_schema(text)
RETURNS text LANGUAGE SQL AS
$func$
SELECT
 table_schema::text
FROM information_schema.tables
 NATURAL JOIN
  (SELECT i AS schema_index, (current_schemas(true))[i] AS table_schema
   FROM generate_series(1, array_upper(current_schemas(true), 1)) g(i)) AS g
WHERE table_name = $1 ORDER BY schema_index ASC LIMIT 1
$func$;

COMMENT ON FUNCTION table_schema(text) IS
'Identify the schema of an un-qualified table reference based on '
'``current_schemas(true)``';
